/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


// xxxxxxxxxxxxxxx   global variable declarations  xxxxxxxxxxxxxxxx

// These variables below can be directly modified by the user on the options form: 
            
var fixed = true;   // Paradigm type: ( true  == "Fixed", false == "PseudoRandom" )

var item_cnt= 15;    // The total number of items in the task. (maximum allowed for fixed=true => 15 else 30)

var rnd_pos = false;  // Position of digits on the screen: fixed (false) or random (true)

var is_ltr = true; // Directionality within each row (digits direction). true == left-right. false == right-left 

var is_before = true;  // Display digits on the screen before of each audio stimulus (true). false => wait for audio stimulus to end 

var repeat_cnt = 3;  // Maximum number of repetitions of the sound stimulus in each item

var delay = 3;   //  Waiting time before repeating the sound stimulus (seconds)

var font_family = 'Georgia, serif';  // user selected font family. canvas ctx.fontFamily for digits

var timer_enabled = false;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 1200;     // Task time in seconds if clock = 1000ms (default). Default value => 20 min

var prg_type = '0';    // type of information in Progress Bar ('0' => Items, '1' => Time, '2' => Nothing) 

var store = false;     // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser.
 

// Other global constants and variables:

const ftype = 'wav';  // default audio file type. if you have different sound files type you can change here. Warning: don't change directory names (path) and all audio files must be same type 

const clock = 1000;  // basic timer resolution in ms. Default = 1s (1000ms)

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var counter = 0;      // timer counter. if (counter >= time) Timer will call theEnd()

var item;     // current item in task. started from 0

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_task = false;    // task in progress or not 

var delay_cnt;     //  time delay (seconds) for play audio. see appTimer() and playSound()

var go_forward = false; // go to next item when current item is done (flag for timer).

var forward_cnt;   //  it gives about 2s of time delay when item is done. works in pair with go_forward. see appTimer()

var stim = [];     // array = for each item: digits content. array dim: [0..item_cnt-1][0..8]

var answer = [];   //  array = for each item: is this digits selected (1) or not selected (0). array dim: [0..item_cnt-1][0..8]

var error = [];    // number of errors for each item. array dim: [0..item_cnt-1]

var point = [];    // number of points for each item. array dim: [0..item_cnt-1]. 1 point = item successfully completed. 0 = item is not completed successfully

var itime=[];      // display time for each item in task (cca ms)

var repeated = [];  // for each item: the number of completed repetitions of the sound stimulus for each item

var rep_begin = [];  // the number of started repetitions of the sound stimulus for each item

var keys = [];     // time order of selected digits for current item
 
var startTime;     // start of display time for current item (0..item_cnt-1)

var rnd_data = [];  // data for fixed == false. see source code for explanation

var rnd_file = [];  // data for fixed == false. see source code for explanation

var loaded = [];   // for each item item: if sound is already loaded => 1. Else 0 (default)


//  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


// init application [create/start Timer, check storage, show start panel..] 
function init() 
{  
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);  
        
  if (isStorage() && storageAvailable('localStorage'))  {
    document.getElementById('store_group').disabled= false;
  }
  else  {
    document.getElementById('store_msg').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline).";
    document.getElementById('store_group').disabled= true;  
    document.getElementById('clr_store').disabled= true;     // IE11 bug  => childrens are not auto disabled
    document.getElementById('store').disabled= true;         // IE11 bug => childrens are not auto disabled
  }              
    
  loadFont();
	
	showStart();
	
}


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{ 
	if (is_task) 
	{ 			 
		let canvas= document.getElementById('canvas');     
		let win_w= window.innerWidth, win_h= window.innerHeight;
		let w= Math.round( 0.95 * win_w );       
		canvas.width= w;
		canvas.style.width= w + 'px'; // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 
		let h_tot= win_h - document.getElementById('footer').offsetHeight;  
		let h= Math.round( 0.95 * h_tot ); 
		canvas.height= h;// https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 
		canvas.style.height= h + 'px';  // The height property has effect only on block-level elements or on elements with absolute or fixed position.			                                 		
		let _left= Math.round( (win_w - w) / 2 );  // center canvas horizontal on screen 
  	canvas.left= _left;
		canvas.style.left= _left + 'px';		
		let _top=  Math.round( (h_tot - h) / 2 );   //  center canvas vertically 
		canvas.top= _top;   
	  canvas.style.top= _top + 'px'; 
    // repaint digits (screen width and/or height is changed). always repaint visual stimulus (digits). don't play audio file
		updateStim(true, false);   
	}
}	


// application timer handler function
function appTimer()  
{ 
  if (is_task)      // else: nop, idle, do nothing 
  { 		         	   			    
    counter++;
    if (counter >= timer)  {     // task time is limited      
		  addTime();
      theEnd();
      return;
    }     
    else if (prg_type == "1")  {   // "1" => Time. display time on Progress Bar  
      var rest=timer-counter;   // this is remaining time 
      var percent= Math.round(100*rest/timer);    
      document.getElementById("progress").setAttribute("style", "width:"+percent+"%");    // !important
      document.getElementById("items_left").textContent= stringTime(rest);    
    }   
    
    if (go_forward)  
    { // go_forward has priority over delay_cnt in this if condition
      forward_cnt++;
      if (forward_cnt >= 2) {   // cca 2s delay
        nextItem();  //go_forward= false;      
      }
      // return;
    }	
    else if (delay_cnt >= 0)  // auditory stimulus were played at least once for current item => audioEnded()
    { 
      delay_cnt++;
      if (delay_cnt >= delay)
      { 
        if ( repeated[item] < repeat_cnt ) { 
          updateStim(is_before,true); // play audio stimulus again (repeat). Response time expired => Triggers a new sound stimulus     
        }
        else  {
          nextItem();
        }
      }
    }    
  }	
}	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== "undefined") {
     return true;               //  localStorage/sessionStorage is supported     
  } 
  else {
    return false;           // no browser Web Storage support
  }  
}


// necessarily because of IE11 and Edge offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	const arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	let sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement("option");
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex=1; // Georgia, serif
	document.getElementById('font_family_msg').style.fontFamily= sel.value; 
}	


// show start panel 
function showStart()  
{      
  document.getElementById("task_section").style.display= "none"; 
  document.getElementById("start").style.display= "block";
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');      
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  //document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  if (document.getElementById('store_group').disabled === false && localStorage.getItem('act-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('act-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' ACT results stored in your web browser';
  }	
  dirty= false;
  document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  
  document.getElementById('options').style.display='none'; 
  document.getElementById('start').classList.add('center-xy');   
  document.getElementById('show_option').style.display ='block';   
  document.getElementById('start_div').style.display='block';        //.removeAttribute('disabled');  
  document.getElementById('hide_option').style.display ='none';   
  //document.getElementById('start_div').style='text-align:center';  // bug?  (firefox auto remove class 'center' when set style.display.none;)
  //document.getElementById('show_option').style='text-align:center';   // bug?    
    
  document.getElementById('instructions').style.display ='block';
       
  refMsg(); 
  
  document.getElementById('instructions').style.display= 'block';	   
}

      
// refresh user changed application Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';    
     
  switch (my_id)
  {
    	       
    case 'fixed':	
			if (e.value == '0')   {
				msg.textContent= 'Fixed paradigm: always the same auditory stimuli (items, tasks). Maximum=15 items';
				document.getElementById('item_cnt').max= 15;             
      }
			else  {
				msg.textContent= 'Auditory stimuli are generated randomly by a set of specific rules. Useful for exercise or test-retest. Maximum=30 items'; 
				document.getElementById('item_cnt').max= 30; 
			}
			if (!document.getElementById('item_cnt').validity.valid)  {  // auto-correct item_cnt to high allowed value
        document.getElementById('item_cnt').value= (e.value == 0) ? 15 : 30;
			}				
			break;

    case item_cnt:        
    case 'repeat_cnt': 
		case 'delay':
		  if (e.validity.valid)  {
				msg.textContent= ''; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
		  break;
        
    case 'rnd_pos':            
			if (e.value == '0')  {
        msg.textContent= 'The digits will always be in the same position on the screen';
        document.getElementById('p_ltr').style.display= 'block';        
      }
      else  {
        msg.textContent= 'The digits will change position on the screen';
        document.getElementById('p_ltr').style.display= 'none'; 
      }        
      break;
    
    case 'is_ltr':     		             
			msg.textContent= e.value == '1' ? 'Within each row the digits will increase the value by 1 in the left-right direction' : 'Within each row the digits will increase the value by 1 in the right-left direction';    
			break;
    
    case 'is_before': 
      msg.textContent= e.value == '1' ? 'The digits on the screen are displayed before the sound stimulus' : 'The digits on the screen appear only after the auditory stimulus is completed';
      break;
	    				                    
    case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes';
        document.getElementById('p_timer').style.display= 'block';
      }
      else  {  
        msg.textContent= 'No'; //'Timer is disabled';
        document.getElementById('p_timer').style.display= 'none'; 
      }       
      break;

    case 'timer':
      if (e.validity.valid)  {
        msg.textContent= stringTime(e.value); 
        msg.removeAttribute('class');
      }
      else  {
        msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
        msg.setAttribute('class', 'error');
      } 
      break;  
    
    case 'prg_type':  
      if (e.value == '0')  {
        msg.textContent= 'It will be displayed how many items are left to the end.';
      }
      else if (e.value == '1')  {
        msg.textContent= 'The remaining time until the end of the task will be displayed. Only when timer is enabled.';
      }
      else  {
        msg.textContent= ''; //'Nothing';
      }
      break; 
           
    case 'store':                                                                                                                                            
      msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
      break;  
      
  }   // switch (my_id)
       
}


// refresh main validator message for user defined OPTIONS (feedback to user)
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check only for validity.valid object in all input type="number" 
function optionsValid()  
{     
  let res= document.getElementById('item_cnt').validity.valid && document.getElementById('repeat_cnt').validity.valid && document.getElementById('delay').validity.valid  && document.getElementById('timer').validity.valid;                                           
  return res;                 
}


// start task (start research) 
function startTask(e)   
{       
  setPar();
  
  stimGenerator(); 
  
  document.getElementById('task_section').style.display= 'block'; 
  document.getElementById('start').style.display= 'none';

	document.getElementById('end').style.display= 'none';    
  document.getElementById('google_wrapper').style.display='none';
  document.getElementById('table_grp').style.display= 'none';
  document.getElementById('canvas').style.display='block';            
  document.getElementById('canvas').disabled= false; 	
  document.getElementById('footer').style.display='block';   
	
  document.getElementById('items_left').textContent= '';
  	
	if (document.getElementById("prg_type").value === "2") {
	  document.getElementById("footer").style.display= "none"; 
	}	
	else  
	{
		document.getElementById('footer').style.display= 'block';     
		document.getElementById('progress').setAttribute('style', 'width:100%');   		
		if (document.getElementById('prg_type').value === '0') {
		  document.getElementById('progress').className='progress-bar progress-bar-info progress-bar-striped';     // items/total
		}
		else if (document.getElementById('prg_type').value === '1')  {
		  document.getElementById('progress').className= 'progress-bar progress-bar-danger progress-bar-striped';  // remaining time   
		} 	
		document.getElementById('progress').setAttribute('style', 'width:0');  // reset progress bar width
	}
      
  counter= 0;       // reset timer counter
  
  startTime= new Date(); 
  
  is_task= true;    // finally 'activate' timer
	
	pageResize();
	
  initItem();
  
	updateStim(is_before, true);       // draw (or not draw) first stimulus and play sound 
   
}


// set application variable parameters in Options form
function setVarPar()  
{ 
                                                                                                   
  fixed = document.getElementById('fixed').value == '0' ? true : false;   // Paradigm type: ( true  == "Fixed", false == "PseudoRandom" )  

  if (document.getElementById('item_cnt').validity.valid)  {
		item_cnt= parseInt(document.getElementById('item_cnt').value, 10);  // see setPar() will check for item_cnt
	}  
  
  rnd_pos= document.getElementById('rnd_pos').value == '1' ? true : false;   // Position of digits on the screen: fixed (false) or random (true)
  
  if (!rnd_pos) {  // ignore is_ltr for rnd_pos= true
    is_ltr= document.getElementById('is_ltr').value == '1' ? true: false;  // Directionality (digits direction within each row). true == left-right. false == right-left	
  }
  
  is_before= document.getElementById('is_before').value == '1' ? true : false;  // Display digits on the screen immediately at the beginning of each audio stimulus or not  
  	
	if (document.getElementById('repeat_cnt').validity.valid)  {
		repeat_cnt= parseInt(document.getElementById('repeat_cnt').value, 10); 
	}
	
  if (document.getElementById('delay').validity.valid) {
		delay= parseInt(document.getElementById('delay').value, 10); 
	}
	
  font_family= document.getElementById('font_family').value;
	
  timer_enabled= (document.getElementById('timer_enabled').checked) ? true: false;  // timer is enabled or disabled
    
  if (document.getElementById('timer').validity.valid) {
		timer= Math.round(parseInt(document.getElementById('timer').value, 10));  // task time (s)
  }
	
	prg_type= document.getElementById('prg_type').value;  // type of information in Progress Bar  
  
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;      // this web application will store (true) or not store (false) results within client web browser
   
}


// stimulus generator function. generate digits
function stimGenerator()  
{   
  if (rnd_pos)  
  {  // if rnd_pos == true => code complete ignore window.is_ltr variable
    for (let i in stim)  
    { // i => item  
      let sel= {}, rnd;  // sel == already selected digits, rnd = random digit in range 1..9  
      for (let j= 0; j < 9; j++)  
      {     
        do  {
          rnd= getRndInt(1,9);
        } while ( typeof sel[rnd] != 'undefined' );
        sel[rnd]=1;        // remember this digit   
        stim[i][j]= rnd;  	
      }
    }    
  }
  else
  { 
    for (let i in stim)  
    { // i => item
      for (let j = 0; j < 9; j++)  {     
        stim[i][j]= j+1;      	
      }
      if (!is_ltr)  
      {
        for (let j = 8; j >= 2; j-=3)  {   // index 8, 5, 2 => swap digits  
          let tmp= stim[i][j];
          stim[i][j]= stim[i][j-2];
          stim[i][j-2]= tmp;            	
        }    
      }
    }    
  }
  
  if (!fixed)  
  {
    // generate Psudorandom paradigm
		const d = ';'  
    rnd_data= [];
    rnd_file=[];
    let sel= {}, rnd, rnd2, rnd3;  // sel == already selected digits, rnd..rndN = random digit in range 1..9  
    let root= 'audio/rnd/';  // path
    
    /*				
      file name	content	items
      rnd1	Select number	0..1
      rnd6	Select two numbers different from	2
      rnd9	Select only one number greater than	3
      rnd10	Select only one number less than	4
      rnd7	Select three numbers different from	5
      rnd11	Select two numbers greater than	6
      rnd12	Select two numbers less than	7
      rnd8	Select two numbers whose sum is not equal	8
      rnd13	Select any two odd numbers whose sum is equal	9
      rnd14	Select any two odd numbers whose sum is not equal	10
      
      rnd2	First select a number	
      rnd3	Then select the number	11
      rnd4	Finally select the number	12
      rnd5	Select the numbers exactly in this order	13..14
    */
    
    
    // item0..item10 generate only 1 random digit
    for (let i= 0; i<= 10; i++)  
    { 
      switch (i)
      {
        case 0: // Select number => item0 and item1
        case 1:
          do  {
            rnd= getRndInt(1,9);
          } while ( typeof sel[rnd] != 'undefined' ); // item0 and item1 must have different numbers
          sel[rnd]=1;        // remember this digit
					break;
          
        // case2 => Select two numbers different from	== item2 => case default
          
        case 3:  // Select only one number greater than => item3             
          rnd= getRndInt(1,8);               
          break; 
        
        case 4:  // Select only one number less than	=> item4            
          rnd= getRndInt(2,9);                
          break;   
 
        //case 5:  Select three numbers different from item5 => default
        
        case 6: // Select two numbers greater than => item6 
			    rnd= getRndInt(1,7);
					break;
				
        case 7:   // Select two numbers less than	=> item7 
          rnd= getRndInt(3,9);                
          break;  
         
        //case 8:  => Select two numbers whose sum is not equal	=> default
         
        case 9: // Select any two odd numbers whose sum is equal	=> item9
          do  {
            rnd= getRndInt(1,9);
          } while ( rnd != 4 && rnd != 6 && rnd != 8 );  // 1 + 3 = 4;  1 + 5 = 6;  1 + 7 = 8;                  
          break;

        case  10: // Select any two odd numbers whose sum is not equal	=> item10 
          rnd= getRndInt(7,9);  // minimum 2 solutions: 1 + 3 = 4;  1 + 5 = 6;  
          break;

        default:
          rnd= getRndInt(1,9);
          //break;
      }
             
      rnd_data[i]= rnd;  // save this digit
      
      rnd_file[i]= []; 
      	
			switch (i)
			{
			  case 0:
				case 1:
					rnd_file[i][0]= root + 'rnd1.'  + ftype;  // Select number
					break;
				
				case 2:
					rnd_file[i][0]= root + 'rnd6.' + ftype; // Select two numbers different from
					break;
        	
        case 3:
					rnd_file[i][0]= root + 'rnd9.' + ftype; // Select any number greater than
					break;	

        case 4:
					rnd_file[i][0]= root + 'rnd10.' + ftype;  // Select any number less than
					break;

        case 5:
					rnd_file[i][0]= root + 'rnd7.' + ftype;  // Select three numbers different from
					break;

        case 6:
					rnd_file[i][0]= root + 'rnd11.' + ftype;  // Select two numbers greater than
					break;

				case 7:
					rnd_file[i][0]= root + 'rnd12.' + ftype;  // Select two numbers less than
					break;
					
				case 8:
					rnd_file[i][0]= root + 'rnd8.' + ftype;  // Select two numbers whose sum is not equal
					break;					

				case 9:
					rnd_file[i][0]= root + 'rnd13.' + ftype;  // Select any two odd numbers whose sum is equal
					break;					
				
				case 10:
					rnd_file[i][0]= root + 'rnd14.' + ftype;  // Select any two odd numbers whose sum is not equal
					break;				
					
			}
			           
      rnd_file[i][1]= root + rnd + '.' + ftype;     // item0..item10 generate only 1 random digit 
    }
   
    // item11=> First select a number + Then select the number => generate 2 random digits
    sel= {};   
		do  {
			rnd=  getRndInt(1,9);
			rnd2= getRndInt(1,9);
		} while ( rnd == rnd2 || typeof sel[rnd] != 'undefined' || typeof sel[rnd2] != 'undefined' );
		sel[rnd]=1;        // remember this 2 digits   
		sel[rnd2]=1;   
		rnd_data[11]= rnd + d + rnd2;  
		rnd_file[11]= [];   
		rnd_file[11][0]= root + 'rnd2.' + ftype;  // rnd2.wav =>	First select a number		
		rnd_file[11][1]= root + rnd + '.' + ftype; // random digit
		rnd_file[11][2]= root + 'rnd3.' + ftype;  // rnd3.wav =>	Then select the number
		rnd_file[11][3]= root + rnd2 + '.' + ftype;			// random digit
    		
		// item12 => First select a number + Then select the number  + Finally select the number => generate 3 random digits
    sel= {}; // generate 3 different digits
		do  {
			rnd= getRndInt(1,9);
		} while (typeof sel[rnd] != 'undefined'); 
		sel[rnd]=1;  // remember 1. digit  
    
		do  {
			rnd2= getRndInt(1,9);
		} while (typeof sel[rnd2] != 'undefined'); 
		sel[rnd2]=1;   // remember 2. digit 

		do  {
			rnd3= getRndInt(1,9);
		} while (typeof sel[rnd3] != 'undefined'); 
		sel[rnd3]=1;   // remember 3. digit 
 		
		rnd_data[12]= rnd + d + rnd2 + d + rnd3;
		rnd_file[12]= []; 
		rnd_file[12][0]= root + 'rnd2.' + ftype; // 	rnd2.wav	=> First select a number	
		rnd_file[12][1]= root + rnd + '.' + ftype;  // add 1. random digit
		rnd_file[12][2]= root + 'rnd3.' + ftype;  // rnd3.wav =>	Then select the number
		rnd_file[12][3]= root + rnd2 + '.' + ftype;		// add 2. random digit
		rnd_file[12][4]= root + 'rnd4.' + ftype;   // rnd4.wav => Finally select the number	
		rnd_file[12][5]= root + rnd3 + '.' + ftype;			// add 3. random digit (last one)	
    
		// items13,item14.. => Select the numbers exactly in this order	(rnd5) started from item13..
    // item13 start with 5 random digits, then code add one digit for next items (item14 = 6 digit. item15 = 7 digit, item16 = 8 digit, item 17 = 9 digit)
    // started from item17 all items are same type (9 digits) 		
		for (let i= 13; i < item_cnt; i++)  
    { 
	    let s='', up;
			rnd_file[i]= [];
			rnd_file[i][0]= root + 'rnd5.' + ftype; // 	rnd5.wav	=> Select the numbers exactly in this order 
			switch (i)  // up => number of random digits for items 13,14,..
			{
				case 13: 
					up= 5;
					break;
					
				case 14:  
					up= 6;
					break;
					
				case 15:  
					up= 7;
					break;
					
				case 16:  
					up= 8;					
					break;
					
				default:  // items: 17,18,.. => started from item17 all items are same type (9 digits) 
					up= 9;
					//break;						
			}		
      sel={}; 			
			for (var j= 1; j<=up; j++)  
			{  // generate 5..9 different digits
				do  {
					rnd= getRndInt(1,9);
				} while ( typeof sel[rnd] != 'undefined' );
				sel[rnd]=1;        // remember this digit   
				rnd_file[i][j]= root + rnd + '.' + ftype;
				s+= rnd + ';';
			}
			rnd_data[i]= s;  // finally add digits (as string)
	  }
    
  }  // if (!fixed) block  
   
}


// set internal application parameters
function setPar()  
{    
  if (item_cnt < 1 || (item_cnt > 15 && fixed))  {
		item_cnt= 15;  // this is probably redundant
	}		
	item= 0;
  stim= [];  
	answer= [];  
	error= [];	
	point= [];
  itime=[];
  repeated= [];
  rep_begin= [];
	loaded=[];
	for (let i= 0; i< item_cnt; i++)   
  { // i => item
    stim[i]= [];  
    answer[i]= [];	
    error[i]= 0;  // number of errors for each item
    point[i]= 0; // 1 point = item successfully completed. 0 = item not completed successfully
    itime[i]=0;  // time (cca ms)
    repeated[i]= 0;
		rep_begin[i]= 0;
    loaded[i]= 0;
		for (let j= 0; j< 9; j++)  {              
			answer[i][j]= 0;   // set all digits within each item as not selected => 0    	
    }
  }   
}


// get random integer from min to max range
function getRndInt(min, max) {
  return Math.floor(Math.random() * (max - min + 1)) + min;
}


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  let hh = Math.floor(total/(60*60));         
  let divisor_for_minutes = total % (60 * 60);
  let mm = Math.floor(divisor_for_minutes / 60);       
  let divisor_for_seconds = divisor_for_minutes % 60;
  let ss = Math.ceil(divisor_for_seconds);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// goto next item (stimulus)
function nextItem()  
{ 
  addTime();
  item++; 
	initItem();	
	if (item < item_cnt)  {         		 
    startTime= new Date();
    updateStim(is_before, true);     // next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


// init item variables
function initItem()  {
  go_forward=false; 
  keys= [];   // reset selected digits for current item
	document.body.style.backgroundColor= '#ffffff';	    
}  


/* update current stimulus
  play_video => draw visual part of stimulus (grid with digits) or not
	play_audio => play sound stimulus or not for current item  */
function updateStim(play_video, play_audio) 
{     
  if (item < 0 || item >= item_cnt)  {
    return;
  }    
  let canvas= document.getElementById('canvas');
  let ctx = document.getElementById('canvas').getContext('2d'); 
  ctx.clearRect(0, 0, canvas.width, canvas.height);      // clear canvas             
  document.getElementById('task_section').style.display= 'none';
  document.getElementById('canvas').disabled= false;  	
	
	if (play_video)  
  {
		// draw (update) digits for current item
		let dx= Math.floor(canvas.width/3);  
		let dy= Math.floor(canvas.height/3); 
		let rad= Math.floor(Math.min(dx,dy)/2);  // text fontSize
		ctx.font= rad.toString() + 'px ' + font_family; 	
		ctx.textBaseline= 'bottom';  
		ctx.textBaseline= 'middle';    
		ctx.textAlign= 'center';
		
		let pos= 0; // pos = digits within current item  
    for (let col= 0; col < 3; col++)  
    {                                              				
      for (let row= 0; row < 3; row++)
      {
        let _left= Math.round(row * canvas.width/3);
        let _top=  Math.round(col * canvas.height/3);
        ctx.beginPath();    
        ctx.fillStyle=  answer[item][pos] ==  '1' ?  '#e5e5e5' : '#ffffff';  // set background color for selected digit within current item 
        let x= _left + dx/2;
        let y= _top + dy/2;
        ctx.lineWidth= 3;
				ctx.rect(_left, _top, dx, dy);   // context.rect(x,y,width,height);                    
        ctx.fill(); 
				ctx.stroke();  				        
				ctx.fillStyle = "#000000";  // change line color into black color (must be)	
        ctx.fillText(stim[item][pos], x, y);
        pos++;
      }		
    }     
    // update Progress bar on footer => show current item count and remaining item count on Progress Bar (default)          
    if (prg_type == '0' )  {                                                                                                                                                 
      let percent = Math.round(100*(item_cnt-item)/item_cnt);      
      document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
      document.getElementById('items_left').textContent= 'Item = ' + (item + 1) + ' / ' + item_cnt;  
    }  
		document.getElementById('task_section').style.display= 'block';  
  }		
      
  if (play_audio) {
    if (typeof rep_begin[item] != 'undefined') {
      rep_begin[item]++; 
    }
    if (fixed)  {
			playSound();   
		}
		else  {
			playMultipleSound(0);
		}
  }
}


// play only one sound stimulus. play audio from ../audio/fix dir. 
function playSound()
{  
	if (typeof loaded[item] === 'undefined') {
    return;
  }  
  delay_cnt= -1;  // timer must wait. audio play in progress.
	let file= 'audio/fix/item' + parseInt(item + 1, 10) + '.'  + ftype ; 
	let sound= document.getElementById('sound');		
	if ( loaded[item] == 0 ) { 
		sound.setAttribute('src', file);  // sound.load();
		loaded[item]=1;				
	}
	sound.play();
	sound.onended = function() {
		audioEnded();    
	}; 			
  
} 


// play a multiple sounds (recursive function). play audio from ../audio/rnd dir. 
function playMultipleSound(cnt)
{ 
  cnt= cnt || 0;  //(typeof(cnt) === "undefined") ? 0 : cnt;
	if ( typeof rnd_file[item] === 'undefined' || typeof rnd_file[item][cnt] === 'undefined' ) {
    return;
  }    
  delay_cnt= -1;  // timer must wait. audio play in progress.			 
  let sound= document.getElementById('sound');		   
	let file= rnd_file[item][cnt];                  
	sound.src= file;   //sound.setAttribute('src', file);  //sound.load(); 
	sound.play();
	if ( cnt == rnd_file[item].length-1 ) 
	{
		sound.onended = function() {
			audioEnded();
		};		
	}
	else 
	{
		sound.onended = function() {
			cnt++;
			playMultipleSound(cnt);    // recursion
		}; 		
	} 			
  	
} 


// audio ended => when the audio file is finished
function audioEnded()
{
  if (!is_before) {
    updateStim(true, false);
  }
  repeated[item]++;
  delay_cnt= 0; 
}  

 	
// user selected digit on canvas => html: <canvas id="canvas" onclick="doAnswer(event)"></canvas>   
function doAnswer(e)  
{   
  let canvas= document.getElementById('canvas');    
  if (canvas.disabled) { 
		return; 
	}
	let x= e.offsetX, y= e.offsetY;  //let x= e.clientX, y= e.clientY; 
  let dx =  canvas.width / 3, dy = canvas.height / 3;  
  let row = Math.floor ( y / dy );  // The offsetX (offsetY) property returns the x (y) -coordinate of the mouse pointer, relative to the target element. 
  let col = Math.floor ( x / dx );  // The clientX (clientY) property returns the horizontal (vertical) coordinate (according to the client area) of the mouse pointer when a mouse event was triggered.  
  let pos = row * 3 + col;   // rectangle position within html canvas object  	
	if ( pos > 8 || pos < 0 )  {  // pos < 0 => probably impossible?
		return;  
  }     
  // set digit is selected or not selected (toogle answer)       
	if (answer[item][pos] == 0)  {   
	  answer[item][pos]= 1;
	  keys.push( stim[item][pos] );
	}
	else  {
	  answer[item][pos]= 0;
    let tmp=[];
		for (let i in keys)  {
		  if ( keys[i] != stim[item][pos] )  {
				tmp.push(keys[i]);
			}
		}			
		keys= tmp;
	}
  
	updateStim(true, false);   // repaint => update backgroundColor for (un)selected digits
	let is_done; 
	if (fixed){
		is_done= itemDone(pos);
  }
  else  {
		is_done = rnditemDone(pos);
  } 
	
	if ( is_done )  {
    document.body.style.backgroundColor= '#2E8B57';  // seagreen
    document.getElementById('canvas').disabled= true;  // tmp disable canvas
		point[item]=1;
    forward_cnt= 0;
    go_forward= true;  // for timer
  }	    
 
}


/* Paradigm type => fixed == true  == "Fixed"
   paradigm fixed. check if current item is done. if is done goto next item. 
   argument: pos => position for last selected digit in current item */
function itemDone(pos)
{
  let digit= stim[item][pos], sel, ok, tmp; // digit => last selected digit (in last selected screen position)
  switch (item)
  {
    case 0:   // Select number 2 
    case 1:  // Select number 7  		
      tmp= (item == 0) ? 2 : 7;
			if ( isOnly(tmp) )  { 
        return true;
      }
      else if ( answer[item][pos] == 1 )  {
        error[item]++;   // increase error only if last digit is selected
      }        
			break;          
      
    case 2:  //  Select two numbers within the same row
      tmp = totalSelected();
			if ( tmp == 2  && ( getSelectedInRange(0,2) == 2 || getSelectedInRange(3,5) == 2 || getSelectedInRange(6,8) == 2 ) )  {							
				return true;								
			}
      else if ( tmp >= 2 && answer[item][pos] == 1 )  {  
				error[item]++;     // increase error only if last digit is selected and >=2 digits are selected
			}			
      break;
      
    case 3:  // Select two numbers. You must not select numbers 7 and 1
      tmp = totalSelected();
			if ( tmp == 2 && ( !isSelected(7) && !isSelected(1) ) )  {						
				return true;				
			}
			else if ( answer[item][pos] == 1 && (tmp > 2 || isSelected(7) || isSelected(1) ) )  {  
				error[item]++;      // increase error only if last digit is selected and >=2 digits are selected				
			}			
      break; 
      
    case 4:  // Select only one number greater than 4 and less than 7 
      if ( totalSelected() == 1 && ( isSelected(5) || isSelected(6) ) )  {
        return true;
      }
      else if ( answer[item][pos] == 1 && digit != 5 && digit != 6 )  {  
        error[item]++;  
      }
      break;

    case 5:  // Select one number not equal to 3 or 8    
      if ( totalSelected() == 1 && !isSelected(3) && !isSelected(8) )  { 	
        return true;
      }  
      else if ( answer[item][pos] == 1 && (digit == 3 || digit == 8)  )  {  
        error[item]++;  
      }  
      break;
      
    case 6:  // Select two different even numbers
      sel= getAllSelected();
      ok = sel.length == 2 ? true : false;
      if (ok) 
      {
        for (let i in sel)  
        {
          if (sel[i] % 2 != 0)  {
            ok= false;
            break;
          }
        }                  
      }
      if (ok)  {  
        return true;
      }
      else if ( answer[item][pos] == 1 && ( sel.length > 2 || digit % 2 != 0 ) )  {  
        error[item]++;         
      }
      break; 
      
      case 7:  // Select two numbers whose sum is 5
				sel= getAllSelected();
				ok = sel.length == 2 ? true : false;      
				if (ok) 
				{
					let sum= 0;
					for (let i in sel)  {
						sum+= sel[i]; 
					} 
					if (sum != 5)  {
						ok= false;
					}          
				}
				if (ok)  {  
					return true;
				}
				else if ( answer[item][pos] == 1 && sel.length >= 2 )  {  
					error[item]++;  
				}
				break; 
                 
      case 8:  // After selecting number 4 select number 7        
				if ( totalSelected() == 2 && ( isSelected(4) && digit == 7 ) )  {					
					return true;										
				}	
        else if ( answer[item][pos] == 1 )  // note: not all errors are catch
        {
					if (digit != 4 && digit != 7 )  { 
						error[item]++;  
					} 
					else if ( !isSelected(4) && digit == 7 ) {  // reverse order
						error[item]++;
					}
					else if ( totalSelected() > 2 ) {  
						error[item]++;
					}					
				}	
        break;

      case 9:   // Before selecting number 9, select number 2
				if ( totalSelected() == 2 && isSelected(2) && digit == 9  ) {  	
					 return true;					
				}				
        else if (answer[item][pos] == 1)  // note: not all errors are catch
        {
					if (digit != 2 && digit != 9 )  { 
						error[item]++;  
					} 
					else if ( isSelected(9) && digit == 2 ) {  // reverse order
						error[item]++;
					}
					else if ( totalSelected() > 2 ) {  
						error[item]++;
					}			
				}
				break;
                
      case 10:  // Select any two odd numbers whose sum is equal 4
        tmp = totalSelected();
				if ( tmp == 2 && isSelected(1) && isSelected(3)  )  {  // only one possible combination is: 1 + 3 = 4
					return true;
				}
				else if ( answer[item][pos] == 1 && ( ( digit != 1 && digit != 3 ) || tmp > 2 )  )  {  
					error[item]++;  
				} 
        break; 
				
			case 11: // Select number 5. Then select number 9. Finally, select number 2
        if ( keys.length == 3 && keys[0] == 5 &&  keys[1] == 9 && keys[2] == 2 )  { 
					return true;
				}
        else if ( answer[item][pos] == 1 && digit != 5 && digit != 9 && digit != 2  )  {  
					error[item]++;  // note: not all errors are catch
				} 
				else if ( answer[item][pos] == 1 && totalSelected() > 3 )  { 
					error[item]++;  // note: not all errors are catch
				} 				
        break; 

      case 12:  // Select three numbers. You must not select numbers 8, 2, 5
        tmp = totalSelected();
				if ( tmp == 3 && !isSelected(8) && !isSelected(2) && !isSelected(5) )  { 
          return true;
        }
				else if ( answer[item][pos] == 1 && ( digit == 8 || digit == 2 || digit == 5 || tmp > 3)  )  {  
					error[item]++; 
				}        				
        break;
 
      case 13:  // Select the numbers exactly in this order 7, 1, 9, 4, 3, 8
        if ( keys.length == 6 && keys[0] == 7 && keys[1] == 1 && keys[2] == 9 && keys[3] == 4 && keys[4] == 3 && keys[5] == 8 )  {
					return true;
				}
        else if ( answer[item][pos] == 1 && ( digit == 2 || digit == 5 || digit == 6 || keys.length > 6)  )  {  
					error[item]++;  
				} 
        break;

      case 14:  // Select two odd numbers that are different from 9,1,5
        if ( totalSelected() == 2 && isSelected(3) && isSelected(7)  )  {  // only one possible combination is: 3 and 7
					return true;
				}
				else if ( answer[item][pos] == 1 && digit != 3 && digit != 7  )  {  
					error[item]++;  // any selected key except 3 and 7 is error. 
				}
				break; 			        
  }
	return false;   // note this line
}  


/* Paradigm type => fixed == false  == "Pseudorandom"
   paradigm fixed. check if current item is done. if is done goto next item. 
   argument: pos => position for last selected digit in current item */
function rnditemDone(pos)
{

	let digit= stim[item][pos], sel, ok, rnd, tmp; // digit => digit in last selected screen position

	/*
	file name	content	items
	rnd1	Select number	0..1
	rnd6	Select two numbers different from	2
	rnd9	Select any number greater than 3
	rnd10	Select any number less than	4
	rnd7	Select three numbers different from	5
	rnd11	Select two numbers greater than	6
	rnd12	Select two numbers less than	7
	rnd8	Select two numbers whose sum is not equal	8
	rnd13	Select any two odd numbers whose sum is equal	9
	rnd14	Select any two odd numbers whose sum is not equal	10
	rnd2	First select a number	
	rnd3	Then select the number	11
	rnd4	Finally select the number	12
	rnd5	Select the numbers exactly in this order	13..14
	*/
	
	switch (item)
  {
    case 0:   // Select number x 
    case 1:		
      if ( isOnly(rnd_data[item]) )  {  
        return true;
      }
      else if ( answer[item][pos] == 1 )  {  
        error[item]++;   // increase error only if last digit is selected
      }  
      break;
     
    case 2:  // Select two numbers different from	
			tmp = totalSelected();
			if ( tmp == 2 && !isSelected(rnd_data[item])   )  {  
				return true;
			}
			else if ( answer[item][pos] == 1 && ( digit == rnd_data[item] || tmp > 2 )  )  {  
				error[item]++;  // forbiden key is selected 				
			}
			break; 				

    case 3:  // Select only one number greater than
		  	sel= getAllSelected();
				ok = sel.length == 1 ? true : false;      
				if (ok) 
				{
					for (let i in sel)  
					{
						if (sel[i] <= rnd_data[item] )  {
              ok= false;
							break;
            } 							
					}        
				}
				if (ok)  {  
					return true;
				}
				else if ( answer[item][pos] == 1 )  {  
					error[item]++;  
				}				
				break;
		
		case 4:  // Select any number less than	
			sel= getAllSelected();
			ok = sel.length == 1 ? true : false;      
			if (ok) 
			{
				for (let i in sel)  
				{
					if (sel[i] >= rnd_data[item] )  {
						ok= false;
						break;
					} 							
				}        
			}
			if (ok)  {  
				return true; 
			}
			else if ( answer[item][pos] == 1 )  {  
				error[item]++;  
			}				
			break;		
		
		case 5:  // Select three numbers different from	
			tmp = totalSelected();
			if ( tmp == 3 && !isSelected(rnd_data[item]) )  {  
				return true;
			}
			else if ( answer[item][pos] == 1 && ( digit == rnd_data[item] || tmp > 3 ) )  {  
				error[item]++;  // forbidden key is pressed or more then 3 digits are selected				
			}
			break; 
		
		case 6:  //	Select two numbers greater than	
			sel= getAllSelected();
			ok = sel.length == 2 ? true : false;      
			if (ok) 
			{
				for (let i in sel)  
				{
					if (sel[i] <= rnd_data[item] )  {
						ok= false;
						break;
					} 							
				}        
			}
			if (ok)  {
				return true;
			}
			else if ( answer[item][pos] == 1 && digit <= rnd_data[item])  {  
				error[item]++;  // note: not all errors are catch
			}				
			break;
		  		
		case 7:  // Select two numbers less than	
			sel= getAllSelected();
			ok = sel.length == 2 ? true : false;      
			if (ok) 
			{
				for (let i in sel)  
				{
					if (sel[i] >= rnd_data[item] )  {
						ok= false;
						break;
					} 							
				}        
			}
			if (ok)  {
				return true;
			}
			else if ( answer[item][pos] == 1 && digit >= rnd_data[item] )  {  
				error[item]++;  // note: not all errors are catch
			}				
			break;
		  
		case 8:  // Select two numbers whose sum is not equal	
			sel= getAllSelected();
			ok = sel.length == 2 ? true : false;      
			if (ok) 
			{
				let sum= 0;
				for (let i in sel)  {
					sum+= sel[i]; 
				} 
				if (sum == rnd_data[item])  {
					ok= false;
				}          
			}
			if (ok)  {
				return true;
			}
			else if ( sel.length >= 2 && answer[item][pos] == 1 )  {  
				error[item]++;  
			}
			break; 
			
		case 9:   // Select any two odd numbers whose sum is equal	
		case 10:  // Select any two odd numbers whose sum is not equal
			sel= getAllSelected();
			ok = sel.length == 2 ? true : false;      
			if (ok) 
			{
				let sum= 0;
				for (let i in sel)  {
					if (sel[i] % 2 == 0)  {
            ok= false;
            break;
          }
					sum+= sel[i]; 
				} 
				if (item == 9) 
				{				// Select any two odd numbers whose sum is equal		
					if (sum != rnd_data[item])  {
						ok= false;
					} 
				}
        else  {  // Select any two odd numbers whose sum is not equal
					if (sum == rnd_data[item])  {
						ok= false;
					}  
        }					
			}
			if (ok)  {
				return true;
			}
			else if ( answer[item][pos] == 1 && ( digit % 2 == 0 || sel.length >= 2)   )  {  
				error[item]++;  
			}
			break; 
				
		case 11:    // 	First select a number. Then select the number		
			rnd= rnd_data[item].split(';'); 
			if ( keys.length == 2 && keys[0] == rnd[0] && keys[1] == rnd[1] )  {
				return true;
			}
			else if ( answer[item][pos] == 1 && digit != rnd[0] && digit != rnd[1]  )  {  
				error[item]++;  
			} 
			else if ( answer[item][pos] == 1 && totalSelected() > 2 )  { 
					error[item]++;  // note: not all errors are catch					
			} 	
			break; 
   
		case 12:    // 	First select a number. Then select the number. Finally select the number	12		
			rnd= rnd_data[item].split(';'); 
			if ( keys.length == 3 && keys[0] == rnd[0] && keys[1] == rnd[1] && keys[2] == rnd[2])  {
				return true;
			}
			else if ( answer[item][pos] == 1 && ( ( digit != rnd[0] && digit != rnd[1] && digit != rnd[2] ) || totalSelected() > 3 ) )  {  
				error[item]++;   // note: not all errors are catch
			} 
			break;
		 
		default:  // Select the numbers exactly in this order: rnd1, rnd2, ... , rnd9
		  rnd= rnd_data[item].split(';'); 			
			let up;
      switch (item)  // up => number of random digits for items 13,14,..
			{
				case 13: 
					up= 5;
					break;					
				case 14:  
					up= 6;
					break;					
				case 15:  
					up= 7;
					break;					
				case 16:  
					up= 8;					
					break;					
				default:  // items: 17,18,.. => started from item17 all items are same type (9 digits) 
					up= 9;
					//break;						
			}					
			if (keys.length == up) 
			{				
	      ok= true;	
				for (var i= 0; i< up; i++)  
				{
					if ( keys[i] != rnd[i] ) {
						ok= false;
						break;
					}
				}
				if (ok)  {
					return true;
				}
				else if ( answer[item][pos] == 1 )  {  
					error[item]++;  // only when any other keys is selected				
			  } 	
			}
      else if ( answer[item][pos] == 1 ) 
			{ 
				ok= false;	
				for (var i= 0; i< up; i++)  
				{
					if ( digit == rnd[i] ) {
						ok= true;
						break;
					}
				}
        if (!ok)  {
					error[item]++;  // last key pressed was not in rnd1..rnd9	
				}
      }				
		 //break;
		 		 
  }
	
	return false;   // note this line
}  


// check if ONLY this digit is selected in current item
// return true if this digit is selected and no other digits are selected within the current item
function isOnly(digit)
{
  for (let j= 0; j< 9; j++)  {  // stim and answer arrays have equal dimensions
    if ( ( stim[item][j] == digit && answer[item][j] == 0 ) || ( stim[item][j] != digit && answer[item][j] == 1 ) ) {
      return false;
    }      
  }    
  return true;
}


// check if this digit is selected in current item
// return true if this digit is selected  (Note: other digits CAN BE selected within the current item)
function isSelected(digit)
{
  for (let j= 0; j< 9; j++)  {  // stim and answer arrays have equal dimensions
    if ( stim[item][j] == digit && answer[item][j] == 1 ) {  
      return true;
    }      
  }    
  return false;
}


// return total number of selected digits within current item
function totalSelected()
{
  let res=0;
  for (let j= 0; j< 9; j++)  {  // stim and answer arrays have equal dimensions
    if ( answer[item][j] == 1 )  {
      res++;
    }      
  }    
  return res;
}


// return number of selected digits in POSITIONS range [low..high] for current item. not look in digit conetnt
function getSelectedInRange(low,high)
{
  let res=0;
  for (let pos= 0; pos< 9; pos++)  {  // stim and answer arrays have equal dimensions
    if ( answer[item][pos] == 1 && pos >= low && pos <= high )  {
      res++;
    }      
  }    
  return res;
}


// return array of all selected digits for current item 
function getAllSelected()
{
  let res= [];
  for (let j= 0; j< 9; j++)  {  // stim and answer arrays have equal dimensions
    if ( answer[item][j] == 1 ) {  
      res.push(stim[item][j]);
    }      
  }    
  return res;
}


// end of task (research)
function theEnd()  
{      
  is_task= false;
  
  document.body.style.backgroundColor= '#ffffff';	
	document.getElementById('task_section').style.display= 'block';
  document.getElementById('canvas').style.display= 'none';   
	document.getElementById('footer').style.display= 'none';
	document.getElementById('end').style.display= 'block';                                                     
  document.getElementById('no_google').textContent='';  
  
  let cor=0, err= 0, mis= 0;
	for (let i in point)   { 
    err+= error[i];  // number of errors for each item
    cor+= point[i]; // 1 point = item is successfully completed. 0 = item not completed successfully
    if (point[i] === 0 && error[i] === 0) {
      mis++;
    }  
  }
  document.getElementById('score').textContent= 'Correct= ' + cor + ' Errors= ' + err + ' Missing:= ' + mis;  
  
  try  {
    drawReport(cor, err, mis);  
  }
  catch (e)  {
    document.getElementById('no_google').textContent= 'You must have an available internet connection to display graphs and tables. Make sure you are online.'; 
  } 
  
  try 
  {
    if (store) {
      document.getElementById('table_grp').style.display= 'block';
      setStorage(cor, err, mis);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').innerHTML= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }
           
}


/* async draw and display tables and charts with user results (call google API).. 
  res = total correct answers in task (sum) => points
  err = total errors in task (sum) => errors
  mis = total missing values in task (sum) => missing */
function drawReport(res, err, mis) 
{          
  document.getElementById('google_wrapper').style.display= 'block';  
  //    xxx  table = Displays the sum of: correct answers, errors, and missing values  xxx 
  let data = new google.visualization.DataTable(), s, elem;    // Create and populate the data table. 
  data.addColumn('string', 'Answer type');     // 1. col = ans array = must be string  
  data.addColumn('number', 'Sum');      // sum (correct answers, errors, missing values) = number = frequencies             
  data.addRow ( ["Correct", res] );
  data.addRow ( ["Errors", err] ); 
  data.addRow ( ["Missing", mis] );                                      
  options= {      
    'width': '50%',    
    'allowHtml': true, 
    'sort': 'enable',  
    'alternatingRowStyle': true              
  };           
  elem= document.getElementById('table');
  elem.style.display = 'block'; 
  //elem.setAttribute('class', 'center_m'); 
  new google.visualization.Table(elem).draw(data, options); 
  //    xxxxxx      PIE chart  => data are same   => draw PIE chart and table for correct, errors, missing values   xxxxxx 
  options= {   
   is3D: true, 
   legend: 'none',
   pieSliceText: 'value',
   tooltip: {text:'value'},    
   width: '100%',
	 title: 'Correct answers (points), errors and missing values',
   chartArea: {width:'100%'}
  };      
  elem = document.getElementById('chart');   
  elem.style.display = 'block'; 
  //elem.setAttribute('class', 'center_m'); 
  new google.visualization.PieChart(elem).draw(data, options);     // Create and draw the visualization.       
    
//  draw table and chart for each item  
//    xxxxxxxxxxxxxx          Chart2 group by stimulus category    xxxxxxxxxxxxxxxxx      

  data = new google.visualization.DataTable();    // Create and populate the data table. 
  data.addColumn('string', 'Item number');     // 1. col = must be string  
  data.addColumn('number', 'Correct'); 
  data.addColumn('number', 'Errors');      
  data.addColumn('number', 'Missing');        
  data.addColumn('number', 'Number of repetitions');
  // rows = items (0..item_cnt-1).  columns = correct(0), errors(1), missing(2)  
	for (let i= 0; i< item_cnt; i++)  {      
    let m= (point[i] === 0 && error[i] === 0) ? 1 : 0;    
    data.addRow([ 'Item ' + parseInt(i+1, 10),  point[i], error[i], m, rep_begin[i] ]);  // data.addRow([ 'Item ' + (i+1), point[i], error[i], {v: 0.1, f: '0'} ]);   
  }              
  options= {
    title: 'Results for all items. Total score = ' + res,
    width: '100%',  
    hAxis: {
      title: 'Item number',
    }, 
    vAxis: {
      title: 'Variables'
    }   
  };    
  elem = document.getElementById('chart2')
  elem.style.display= 'block';   
  //elem.setAttribute('class', 'center_m'); 
  new google.visualization.ColumnChart(elem).draw(data, options);     // Create and draw the visualization.                       
  
   //    xxxxxxxxxxxxxx          Frequencies table2 (data are same)   xxxxxxxxxxxxxxxxx                             
  options= {      
    'width': '50%',    
    'allowHtml': true, 
    'sort': 'enable',  
    'alternatingRowStyle': true              
  };    // data are same         
  elem= document.getElementById('table2');
  elem.title= 'Total score= ' + res;
  elem.style.display = 'block'; 
  //elem.setAttribute('class', 'center_m'); 
  new google.visualization.Table(elem).draw(data, options); 
		
	//  draw table and chart for item time for each stimulus  
  //    xxxxxxxxxxxxxx          chart3 item time for each stimulus xxxxxxxxxxxxxxxxx      

  data = new google.visualization.DataTable();    // Create and populate the data table. 
  data.addColumn('string', 'Item');     // 1. col = must be string  
  data.addColumn('number', 'Time (ms)');        
  for (let i= 0; i< item_cnt; i++)  {                                    		
		data.addRow([ 'Item ' + parseInt(i+1, 10), itime[i] ]);     //  if (itime[i]== 0) data.addRow([ s,  {v: 0.1, f: '0'} ]);       
  }                 
  options= {
    title: 'Time ms for each item. See Note',
    width: '100%',  
    hAxis: {
      title: 'Item number',
    }, 
    vAxis: {
      title: 'Time (ms)'
    }   
  };    
  elem= document.getElementById('chart3');
  elem.style.display = 'block';   
  //elem.setAttribute('class', 'center_m'); 
  new google.visualization.LineChart(elem).draw(data, options);     // Create and draw the visualization.                       
        
  //    xxxxxxxxxxxxxx     Frequencies table3 item time for each stimulus (data are same)   xxxxxxxxxxxxxxxxx                             
  options= {      
    'width': '50%',    
    'allowHtml': true, 
    'sort': 'enable',  
    'alternatingRowStyle': true              
  };    // data are same         
  elem= document.getElementById('table3');
  elem.title= 'Total score= ' + res;
  elem.style.display = 'block'; 
  //elem.setAttribute('class', 'center_m'); 
  new google.visualization.Table(elem).draw(data, options); 

}


/*  use window.localStorage in client browser
    cor == total number of points (1 point for each item if this item is ok done)
    err = total number of errors in task
    mis == total number of missing items in task */  
function setStorage(cor, err, mis)   
{    
  let oldItems = JSON.parse(localStorage.getItem("act-v1.0")) || []; 
  let id= oldItems.length+1, tmp= '';   
  if (!rnd_pos)  { // recode window.is_ltr
    tmp= is_ltr ? 1 : 0;
  }
  // set new "record" 
  let newItem = {
    'id' : id,
    'date' : setDate(),   //new Date(),      
    'correct' : cor, 
    'errors' : err,       
    'missing' : mis,    
    'items' : item_cnt,        
    'paradigm' : fixed ? "F" : "R",
    'position' : rnd_pos ? 'random' : 'fixed',
    'ltr' : tmp,
    'before' : is_before ? 1 : 0,
    'repetitions' : repeat_cnt,
    'delay' : delay,
    'timer' : timer_enabled ? document.getElementById("timer").value : ''             
  };

  oldItems.push(newItem);

  localStorage.setItem("act-v1.0", JSON.stringify(oldItems));

  createTableJSON(document.getElementById("table_grp"), oldItems); 
   
  document.getElementById("score").textContent+= ". (ID=" + id + ")"; 
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  table.setAttribute("class", "my_table");    // table.className = 'my_table';
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


// clear web browser storage data
function clrStorage()  
{  
  let cnt= localStorage.getItem("act-v1.0")  ?  JSON.parse(localStorage.getItem("act-v1.0")).length : 0;                                                        
  if (cnt === 0)  {
    alert("You have no ACT data to delete");
    return;
  }  
  if ( !window.confirm("You have a total of " + cnt  + " ACT results stored in your web browser. Continue?") ) { 
    return;
  }                                             
  if (window.confirm("Delete total saved ACT results:  "  + cnt + ". This action cannot be undone. Final confirmation?") ) 
  { 
   window.localStorage.removeItem("act-v1.0"); 
   document.getElementById("store_cnt").textContent= "All data has been deleted. No data";
  }  
}


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
}	


// add time to current item in task
// see:  https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date/getTime
function addTime()  {
  if (item < item_cnt)  {
    var endTime = new Date(); 
    itime[item]+= endTime.getTime() - startTime.getTime();  // ms  (item == current item in task)	
  }
}	

// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}
